@@ -4,27 +4,16 @@ from typing import Optional
 from .transformer import TransformerEncoder, FeedforwardBlock
 from .cnn import SpaceToDepth, calculate_output_spatial_size, spatial_tuple
 from .activation import ReLU, SquaredReLU, GELU, SwiGLU
+from .utils import PadTensor
 
 from einops import einsum
 from einops.layers.torch import Rearrange
 
+import torch
 import torch.nn as nn
 import torch.nn.functional as F
 
 
-class PadTensor(nn.Module):
-    def __init__(self, *args, **kwargs):
-        super().__init__()
-        self.args = args
-        self.kwargs = kwargs
-
-    def forward(self, x):
-        if sum(self.args[0]) == 0:
-            return x
-        else:
-            return F.pad(x, *self.args, **self.kwargs)
-
-
 class GetCLSToken(nn.Module):
     def __init__(self):
         super().__init__()
@@ -44,10 +33,18 @@ class SequencePool(nn.Module):
             ]
         )
 
+        self.reset_parameters()
+
     def forward(self, x):
         weights = self.attention(x)
         return einsum(weights, x, "batch seq, batch seq d_model -> batch d_model")
 
+    def reset_parameters(self):
+        # Iterate over modules in the sequential block
+        for module in self.attention:
+            if hasattr(module, "reset_parameters"):
+                module.reset_parameters()
+
 
 class ClassificationHead(nn.Module):
     """
@@ -65,10 +62,11 @@ class ClassificationHead(nn.Module):
         self.d_model = d_model
         self.summarize = GetCLSToken()
 
-        if logit_projection_layer is not None:
-            self.projection = logit_projection_layer(d_model, n_classes)
-        else:
+        if d_model == n_classes:
+            # No need to project
             self.projection = nn.Identity()
+        else:
+            self.projection = logit_projection_layer(d_model, n_classes)
 
         if batch_norm_logits:
             self.batch_norm = nn.BatchNorm1d(n_classes, affine=False)
@@ -83,9 +81,16 @@ class ClassificationHead(nn.Module):
             ]
         )
 
+        self.reset_parameters()
+
     def forward(self, x):
         return self.classification_process(x)
 
+    def reset_parameters(self):
+        for module in self.classification_process:
+            if hasattr(module, "reset_parameters"):
+                module.reset_parameters()
+
 
 class SequencePoolClassificationHead(ClassificationHead):
     """
@@ -118,6 +123,8 @@ class SequencePoolClassificationHead(ClassificationHead):
             ]
         )
 
+        self.reset_parameters()
+
 
 class ViTEncoder(nn.Module):
     """
@@ -172,6 +179,7 @@ class ViTEncoder(nn.Module):
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
+        transformer_checkpoint_ff=True,
         linear_module=nn.Linear,
     ):
         super().__init__()
@@ -333,6 +341,7 @@ class ViTEncoder(nn.Module):
                 pre_norm=transformer_pre_norm,
                 normformer=transformer_normformer,
                 post_norm=transformer_post_norm,
+                checkpoint_ff=transformer_checkpoint_ff,
             )
         else:
             self.transformer = nn.Identity()
@@ -366,6 +375,7 @@ class ViTEncoder(nn.Module):
                 normformer=transformer_normformer,
                 post_norm=transformer_post_norm,
                 residual_path=transformer_initial_ff_residual_path,
+                checkpoint=transformer_checkpoint_ff,
             )
         else:
             self.initial_ff = nn.Identity()
@@ -385,9 +395,20 @@ class ViTEncoder(nn.Module):
             ]
         )
 
+        self.reset_parameters()
+
     def forward(self, x):
         return self.encoder(x)
 
+    def attention_scores(self, x):
+        x = self.encoder[:-1](x)
+        return self.encoder[-1].attention_scores(x)
+
+    def reset_parameters(self):
+        for module in self.encoder:
+            if hasattr(module, "reset_parameters"):
+                module.reset_parameters()
+
 
 class ViT(nn.Module):
     """
@@ -438,6 +459,7 @@ class ViT(nn.Module):
         transformer_mlp_dropout=0.0,
         transformer_msa_dropout=0.1,
         transformer_stochastic_depth=0.1,
+        transformer_checkpoint_ff=True,
         head=SequencePoolClassificationHead,
         batch_norm_logits=True,
         logit_projection_layer=nn.Linear,
@@ -504,6 +526,7 @@ class ViT(nn.Module):
             transformer_mlp_dropout=transformer_mlp_dropout,
             transformer_msa_dropout=transformer_msa_dropout,
             transformer_stochastic_depth=transformer_stochastic_depth,
+            transformer_checkpoint_ff=transformer_checkpoint_ff,
             linear_module=linear_module,
         )
 
@@ -514,9 +537,26 @@ class ViT(nn.Module):
             batch_norm_logits=batch_norm_logits,
         )
 
+        self.reset_parameters()
+
     @property
     def sequence_length(self):
         return self.encoder.sequence_length
 
     def forward(self, x):
         return self.pool(self.encoder(x))
+
+    def attention_scores(self, x):
+        return self.encoder.attention_scores(x)
+
+    def head_to_bos_token_attention(self, x):
+        all_attention = self.attention_scores(x)
+        batch_averages = torch.mean(all_attention, dim=0, keepdim=False)
+        sequence_averages = torch.mean(batch_averages, dim=-1, keepdim=False)
+        n_bos_tokens = self.encoder.encoder._bos_tokens
+        just_bos = sequence_averages[:, :, :n_bos_tokens]
+        return F.softmax(just_bos, dim=-1)  # (layer, head, bos_token)
+
+    def reset_parameters(self):
+        self.encoder.reset_parameters()
+        self.pool.reset_parameters()
